from mfnet import NMFnet
import torch
import dgl
import sys
import numpy as np
from tqdm import tqdm
import torch
import torch.nn as nn
import torch.optim as optim
from dgl.data.gindt import GINDataset
from dataloader import GraphDataLoader, collate
from args import args
import torch.nn.functional as F
def init_mf(n,bond,dim3):
    torch.manual_seed(args.seed1)
    marg_i=F.softmax(torch.ones(n, bond, dim3), dim=1)
    return marg_i

def train(args, net, trainloader, optimizer, criterion, epoch):
    net.train()
    running_loss = 0
    total_iters = len(trainloader)
    # setup the offset to avoid the overlap with mouse cursor
    #bar = tqdm(range(total_iters), unit='batch', position=2, file=sys.stdout)
    #for pos, (graphs, labels) in zip(bar, trainloader):
    for graphs,labels in trainloader:
        # batch graphs will be shipped to device in forward part of model
        labels = labels.to(args.device)
        #feat = graphs.ndata.pop('attr').to(args.device)
        feat = init_mf(len(graphs),args.bond,args.channel).to(args.device) 
        #print(graphs)
        #graphs.batch_num_nodes()
        #batch_n=graphs.batch_num_nodes().to(args.device)
        graphs = graphs.to(args.device)
        outputs = net(graphs, feat,args.num_layer)
        loss = criterion(outputs, labels)
        running_loss += loss.item()
        # backprop
        optimizer.zero_grad()
        loss.backward(retain_graph=True)
        optimizer.step()
        # report
        #bar.set_description('epoch-{}'.format(epoch))
    #bar.close()
    # the final batch will be aligned
    running_loss = running_loss / total_iters
    return running_loss
def eval_net(args, net, dataloader, criterion):
    net.eval()
    total = 0
    total_loss = 0
    total_correct = 0
    for data in dataloader:
        graphs, labels = data
        #batch_n=graphs.batch_num_nodes().to(args.device)
        #feat = graphs.ndata.pop('attr').to(args.device)
        feat = init_mf(len(graphs),args.bond,args.channel).to(args.device) 
        graphs = graphs.to(args.device)
        labels = labels.to(args.device)
        total += len(labels)
        outputs = net(graphs, feat, args.num_layer)
        _, predicted = torch.max(outputs.data, 1)
        total_correct += (predicted == labels.data).sum().item()
        loss = criterion(outputs, labels)
        # crossentropy(reduce=True) for default
        total_loss += loss.item() * len(labels)
    loss, acc = 1.0*total_loss / total, 1.0*total_correct / total
    net.train()
    return loss, acc
def main(args):
    # set up seeds, args.seed supported
    torch.autograd.set_detect_anomaly(True)
    torch.manual_seed(seed=args.seed)
    np.random.seed(seed=args.seed)
    is_cuda = not args.disable_cuda and torch.cuda.is_available()
    if is_cuda:
        args.device = torch.device("cuda:" + str(args.device))
        torch.cuda.manual_seed_all(seed=args.seed)
    else:
        args.device = torch.device("cpu")
    dataset = GINDataset(args.dataset, False)
    trainloader, validloader = GraphDataLoader(
        dataset, batch_size=args.batch_size, device=args.device,
        collate_fn=collate, seed=args.seed, shuffle=True,
        split_name='fold10', fold_idx=args.fold_idx).train_valid_loader()
    # or split_name='rand', split_ratio=0.7
    weight=args.beta*torch.exp(torch.randn(args.bond, args.bond, args.channel)).to(args.device)
    model = NMFnet(weight,dataset.gclasses).to(args.device)
    criterion = nn.CrossEntropyLoss()  # defaul reduce is true
    optimizer = optim.Adam(model.parameters(), lr=args.lr)
    scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=50, gamma=0.5)
    
    for epoch in range(args.epoch):
       train(args, model, trainloader, optimizer, criterion, epoch)
       scheduler.step()
       train_loss, train_acc = eval_net(
            args, model, trainloader, criterion)
       valid_loss, valid_acc = eval_net(
            args, model, validloader, criterion)
       
       if not args.filename == "":
            with open(args.filename, 'a') as f:
                f.write('%s' % (
                    args.dataset
                ))
                f.write("\n")
                f.write("%f %f %f %f" % (
                    train_loss,
                    train_acc,
                    valid_loss,
                    valid_acc
                ))
                f.write("\n")     
if __name__=="__main__":
    main(args)            
        
       
