"""
Tensor network class
TODO:
    1. estimate error made in SVD
    2. remove graphx dependences in Tensor_Network class

Long_term TODO:
    1. Opeartions of neighbor should be put to node class. In the current implementations, move(), move2tail() do not operate on neighbors. 
    2. Based on TODO.1, move2tail() and move2head() can be optimized by reversing MPSnode.mps. So far it would be a mess to implement the reverse() in MPSnode.mps. because the MPSnode.neighbor opeartions are implemented outside the class.
"""
import torch
import numpy as np
import math
import networkx as nx
from scipy.linalg import sqrtm
import time
import sys
import raw_node
import mps_node
sys.path.append('..')



class Tensor_Network:
    """ Tensor network for the graphical model.
    Storage: The data are stored in a dictionary *tensors*, each of which is a Node class.
    """

    def __init__(self, n, edges,choice,bond,tensor1,tensor2, beta, seed=1, mydevice='cpu', maxdim=30, verbose=-1, Dmax=1024,
                 chi=32, node_type="raw"):
        self.n = n
        self.G = nx.MultiGraph()
        self.node_type = node_type
        self.chi = chi
        self.device = mydevice
        self.verbose = verbose
        np.random.seed(seed)
        self.maxdim = maxdim
        self.Dmax = Dmax  # maximum bond dimension
        self.cutoff = 1.0e-15
        self.G.add_nodes_from(np.arange(self.n))
        self.G.add_edges_from(edges)
        self.m = len(self.G.edges)
        self.maxdim_intermediate = -1
        self.edges=edges
        self.beta=beta
        max_degree = max(np.array(self.G.degree)[:, 1])
        self.num_isolated = sum((np.array(self.G.degree)[:, 1]) == 0)
        print("maximum degree=", max_degree, ", number of isolated nodes=", self.num_isolated)
        self.choice=choice
        self.bond=bond
        self.ini_tensor1=tensor1
        self.ini_tensor2=tensor2
        self.tensors = []
        self.construct_tensor()
        self.select_edge_init()
    def construct_tensor(self):
        if self.choice==1:
           for key in range(self.n):
               if self.node_type == "raw":
                self.tensors.append(raw_node.Node(torch.exp(self.beta*self.ini_tensor1[self.G.degree(key)-1]), key, []))
        elif self.choice==2:
           for key in range(self.n):
               if self.node_type == "raw":
                tensor=self.get_identensor(self.G.degree(key),self.bond,self.device)
                self.tensors.append(raw_node.Node(tensor, key, []))
        if self.choice==2:
           for edge in range(len(self.edges)):
            i, j = self.edges[edge]
            B=torch.exp(self.beta*self.ini_tensor2)
            Q = B
            R = torch.eye(B.shape[1],dtype=torch.float64,device=self.device)
            if self.node_type == "raw":
                nodei = self.tensors[i]
                idx = len(nodei.neighbor)
                shapei = [a for a in nodei.tensor.shape]
                shapei[-1]=Q.shape[1]
                #print(nodei.tensor)
                #print(Q)
                nodei.tensor = (nodei.tensor.reshape(-1, Q.shape[0]) @ Q).reshape(shapei)
                if idx < nodei.order() - 1:
                    nodei.tensor = nodei.tensor.permute([a for a in range(1, nodei.order())] + [0])
                nodei.neighbor.append(j)
                nodej = self.tensors[j]
                shapej = [a for a in nodej.tensor.shape]
                shapej[-1]=R.shape[1]
                seq = [a for a in range(nodej.order())]
                idx = len(nodej.neighbor)
                x=nodej.tensor.reshape(-1, R.shape[0]) @ R
                
                nodej.tensor = (x).reshape(shapej)
                if idx < nodej.order() - 1:
                    nodej.tensor = nodej.tensor.permute([a for a in range(1, nodej.order())] + [0])
                nodej.neighbor.append(i)
        else:
           for edge in range(len(self.edges)):
            i, j = self.edges[edge]
            if self.node_type == "raw":
                nodei = self.tensors[i]
                nodei.neighbor.append(j)
                nodej = self.tensors[j]
                nodej.neighbor.append(i)
    def get_identensor(self,order,multi,mydevice):
        #print(multi)
        #print(order)
        tensor = torch.zeros(multi ** order, dtype=torch.float64, device=mydevice)
        if multi==1:
           distance=0
        else:
           distance=int((multi**order-1)/(multi-1))
        for  i in range(multi):
             tensor[i*distance] = 1
        
        tensor=tensor.reshape([multi] * order)
        return tensor
            
    def dim_after_merge(self, i, j):
        nodei = self.tensors[i]
        nodej = self.tensors[j]
        #print(nodei.neighbor)
        #print(nodej.neighbor)
        #print(nodei.tensor)
        #print(nodej.tensor)
        idx_j_in_i = nodei.find_neighbor(j)
        #print(idx_j_in_i)
        di = nodei.logdim()
        dj = nodej.logdim()
        d = nodei.logdim(idx_j_in_i)
        return round(di + dj - d * 2)

    def select_edge(self):
        count = min([i if len(self.edge_count[i]) > 0 else math.inf for i in self.edge_count.keys()])
        if count > self.maxdim_intermediate:
            self.maxdim_intermediate = count
        if count > self.maxdim and self.node_type == "raw":
            i, j = self.edge_count[count][0]
            print("Tring to contract tensor", i, "and tensor", j, "intermediate tensor dimension", count)
            nodei = self.tensors[i]
            nodej = self.tensors[j]
            print(i, nodei.shape(), nodei.neighbor)
            print(j, nodej.shape(), nodej.neighbor)
            print("The intermediate tensor is larger than maximum dimension")
            self.print_all_tensor_shape()
            sys.exit(1)

        return self.edge_count[count][0]

    def count_add_nodes(self, nodes):
        edges = []
        for i in nodes:
            edges = edges + [tuple(sorted([i, j])) for j in self.tensors[i].neighbor]
        self.count_add_edges(set(edges))

    def count_add_edges(self, edges):
        """ Notice that two end nodes of each edge should sorted, and edges should be unique """
        for i, j in edges:
            count = self.dim_after_merge(i, j)
            if count in self.edge_count.keys():
                self.edge_count[count].append(sorted([i, j]))
            else:
                self.edge_count[count] = [sorted([i, j])]

    def select_edge_init(self):
        self.edge_count = {}
        self.count_add_edges(set([tuple(sorted(a)) for a in self.G.edges()]))

    def count_remove_nodes(self, nodes):
        for j in nodes:
            for i in self.tensors[j].neighbor:
                count = self.dim_after_merge(i, j)
                if sorted([i, j]) in self.edge_count[count]:
                    self.edge_count[count].remove(sorted([i, j]))

    def print_all_tensor_shape(self):
        for i in range(len(self.tensors)):
            if len(self.tensors[i].tensor) != 0:
                print(i, self.tensors[i].tensor.shape, self.tensors[i].neighbor)

    def contraction(self):
        # self.lnZ = math.log(2) * self.num_isolated
        self.lnZ = torch.log(torch.tensor([2], dtype=torch.float64, device=self.device)) * self.num_isolated
        t_select = 0
        t_contract = 0
        t_svd = 0
        while self.G.number_of_edges() > 0:
            t0 = time.time()
            i, j = self.select_edge()
            if self.tensors[j].order() > self.tensors[i].order():
                i, j = j, i  # this is to ensure that node i has larger degree than node j

            self.count_remove_nodes([i, j] + list(self.tensors[i].neighbor) + list(self.tensors[j].neighbor))
            t_select += time.time() - t0
            neigh1 = self.tensors[i].neighbor
            neigh2 = self.tensors[j].neighbor
            idx_j_in_i = np.argwhere(neigh1 == j)[0][0]
            idx_i_in_j = np.argwhere(neigh2 == i)[0][0]

            t1 = time.time()
            self.tensors[i].delete_neighbor(j)
            duplicate = []
            for l in range(len(neigh2)):
                # arrange neighbors
                if l != idx_i_in_j:
                    k = neigh2[l]
                    idx_k_in_i = self.tensors[i].find_neighbor(k)
                    self.tensors[i].add_neighbor(k)  # append the new neighbor to the neighbor list
                    self.G.add_edge(i, k)
                    idx_i_in_k = self.tensors[k].find_neighbor(i)
                    idx_j_in_k = self.tensors[k].delete_neighbor(j)
                    self.tensors[k].add_neighbor(i, idx_j_in_k)  # add i to k's neighbor list, replaceing j
                    if idx_i_in_k > -1:  # i already in k
                        duplicate.append(k)
                        self.tensors[k].merge(i, cross=idx_i_in_k > idx_j_in_k)
            old_shapei = self.tensors[i].shape()
            old_shapej = self.tensors[j].shape()
            lognorm = self.tensors[i].eat(self.tensors[j], idx_j_in_i, idx_i_in_j)
            self.lnZ += lognorm
            for k in duplicate:
                self.tensors[i].merge(k, cross=False)
                idx_k_in_i = self.tensors[i].find_neighbor(k)
                if self.node_type == "mps" and self.tensors[i].mps[idx_k_in_i].shape[1] > self.Dmax:
                    self.cut_bondim(i, idx_k_in_i)
            self.tensors[j].clear()
            self.G.remove_node(j)
            t_contract += time.time() - t1
            t0 = time.time()

            edges = np.array(list(self.G.edges))
            m_left = 0
            if len(edges) > 0:
                edges = edges[:, :2]
                m_left = len(np.unique(edges, axis=0))

            if m_left > 2 and self.Dmax > 0:
                self.low_rank_approx_site(i)

            t_svd += time.time() - t0
            '''
            if self.verbose > 3:
                print(m_left, "/", self.m, list(old_shapei), "idx", idx_j_in_i, list(old_shapej), "\n\t\t\t\t---->",
                      list(self.tensors[i].shape()))
            if self.verbose > 2:
                print(m_left, "/", self.m, "(%d,%d)" % (i, j), ")", "select %.2f" % t_select,
                      "contact %.2f" % t_contract, "rank_appr. %.2f" % t_svd, self.tensors[i].shape(),
                      "%.2f" % (time.time() - t1), "Sec.")
            else:
                print(m_left, "/", self.m, "(%d,%d)" % (i, j), "\t--->",
                      ",".join([str(a) for a in self.tensors[i].shape()]), "\t%.2f" % (time.time() - t1), "Sec.")
            '''
            self.count_add_nodes([i] + list(self.tensors[i].neighbor))
        lognorm = self.lognorm()
        self.lnZ = self.lnZ + lognorm
        return self.lnZ

    def lognorm(self):
        lognorm = torch.tensor(0, dtype=torch.float64, device=self.device)
        for i in range(self.n):
            lognormi = self.tensors[i].lognorm()
            lognorm = lognorm + lognormi
        return lognorm

    def low_rank_approx(self):
        for i in range(self.n):
            self.low_rank_approx_site(i)

    def low_rank_approx_site(self, i):
        """ Try to do low-dimensional approximations to large bond fo site i"""
        if not self.tensors[i].shape():
            return
        t = self.tensors[i]
        try:
            if t.order() == 0:
                return
        except:
            print("error in low_rank_approximate(", i, ")", "tensor")
            print(t.tensor)
            sys.exit(2)

        while max(self.tensors[i].shape()) > self.Dmax:
            self.cut_bondim(i, np.array(self.tensors[i].shape()).argmax())

    
    def cut_bondim(self,i,idx_j_in_i):
        error = 0
        j=self.tensors[i].neighbor[idx_j_in_i]
        idx_i_in_j = self.tensors[j].find_neighbor(i)
        sys.stdout.flush()
#        print("cutting bond",i,j,"idx",idx_j_in_i,idx_i_in_j)
        if(self.node_type == "raw"):
            mati = self.tensors[i].unfolding(idx_j_in_i)
            matj = self.tensors[j].unfolding(idx_i_in_j)
            merged_matrix = mati@matj.t()
        else:
            da_l = self.tensors[i].mps[idx_j_in_i].shape[0]
            da_r = self.tensors[i].mps[idx_j_in_i].shape[2]
            d = self.tensors[i].mps[idx_j_in_i].shape[1]

            db_l = self.tensors[j].mps[idx_i_in_j].shape[0]
            db_r = self.tensors[j].mps[idx_i_in_j].shape[2]

            mati = self.tensors[i].mps[idx_j_in_i].permute([0,2,1]).reshape(-1,d)

            matj = self.tensors[j].mps[idx_i_in_j].permute([0,2,1]).reshape(-1,d)
            merged_matrix = mati@matj.t()
            merged_matrxi = torch.einsum("ijk,ajb->ikab",self.tensors[i].mps[idx_j_in_i],self.tensors[j].mps[idx_i_in_j]).reshape(da_l*da_r,db_l*db_r)
        try:
            [U,s,V] = svd(merged_matrix)
        except:
            print("SVD failed: shape of merged_matrix",merged_matrix.shape)
            sys.exit(-1)
        s_eff = s[s>self.cutoff]
        myd = min(len(s_eff),self.Dmax)
        if(myd == 0):
            print("Warning: encountered ZERO matrix in cut_bondim()")
            myd = 1
            mati=(U[:,0]*s[0])[:,None]
            matj = ((s[0]*V[:,0].t()).t())[:,None]
        else:
            error = error + s[myd:].sum()
            s_eff=s_eff[:myd]
            s=torch.diag(torch.sqrt(s_eff))
            U=U[:,:myd]
            V=V[:,:myd]
            mati=U@s
            matj = (s@V.t()).t()
        if(self.node_type == "raw"):
            self.tensors[i].restore_from_matrix(mati,idx_j_in_i)
            self.tensors[j].restore_from_matrix(matj,idx_i_in_j)
        else:
            mati = mati.reshape(da_l,da_r,mati.shape[1]).permute([0,2,1])
            self.tensors[i].mps[idx_j_in_i] = mati
            matj = matj.reshape(db_l,db_r,matj.shape[1]).permute([0,2,1])
            self.tensors[j].mps[idx_i_in_j] = matj

        #print(list(self.tensors[i].mps[idx_j_in_i].shape),list(self.tensors[j].mps[idx_i_in_j].shape));
        return error


